/* Language identification contestant
 * Doug Beeferman (baseline code by Chris Manning)
 * 5 Feb 1996
 * Uses hash table presence-based language model
 */


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "sih.h"
#include "general.h"

#define SIH_KEY(STRING,KEY) {\
    int ii = 0; \
    char *pstr = STRING; \
    KEY = 0; \
    do {KEY += *pstr++ << (0xF & --ii);} while (*pstr);  /* Fil's hash function */ \
}


#define CHARSET 256
#define DEBUG 0
#define MAX_N 5
#define MAX_LANGS 1000
#define BUFSIZE 1000
#define BUFLIM 750 
#define USE_PROBABILISTIC 0

double rgBlockThresholds[MAX_N] = {0.0,   0.04,   0.05,   0.06,   1.0,   0.5};
int rgWeights[MAX_N+1] = {-1,   7,   8,   4,  50,  50};
double rgCurrentThresholds[MAX_N+1];

typedef struct lgstruct {
  struct lgstruct *next;
  char name[30];
  char filename[80];

  sih_t *hashCounts;
  int cChars;
} lginfo, *lginfoptr;


void initlginfo(lginfoptr lg);
void dump_buf(lginfoptr lg, char *pBuf, int iHigh);
void print_hash_table(lginfoptr lg);
void print_train_info(lginfoptr lg);
void initlginfo(lginfoptr lg);
lginfoptr calculate_best(lginfoptr lg, char *string);




void perror2(char *template, char *parm)

{
  fprintf(stderr, template, parm);
  exit(0);
}


void processfile(lginfoptr lg)

{
  FILE *infile;
  int c,lastc='\0';
  int i, j;
  char pBuf[BUFSIZE];  /* Circular buffer */
  

  initlginfo(lg);

  if ( ! (infile = fopen(lg -> filename, "r")))
    perror2("Processfile: Can't open file %s.\n", lg -> filename);

  /* i points to current insertion point in buffer,
     j points to beginning of undumped data */
  i=0;
  while ((c = fgetc(infile)) != EOF)
    {
      /* Treat newlines as spaces */
      if (c == '\n') c=' ';
      /* if ((c == ' ') && (lastc == ' ')) continue; */

      pBuf[i] = c;
      i++;
      lg->cChars++;
      if (i > BUFLIM) 
	{
	  /* Dump block */
	  dump_buf(lg, pBuf, i-MAX_N+1);
	  memcpy(pBuf, pBuf+i-MAX_N+1, (MAX_N-1) * sizeof(char));
	  i = MAX_N;
	}
      lastc=c;
    }

  /* Dump whatever remains */
  if (i <= BUFLIM) dump_buf(lg, pBuf, i-MAX_N+1);

  /* print_hash_table(lg); */
  close(infile);
}


void dump_buf(lginfoptr lg, char *pBuf, int iHigh)
{
  int i, j, len;
  char cTmp, *pTmp, *pEnd;
  int iCount;
  sih_t *ht = lg->hashCounts;

  /* Add all (1..MAX_N)-grams to the hash table */
  for (i=0; i<iHigh; i++)
    {
      pTmp = &pBuf[i];
      for (j=1; j<=MAX_N; j++)
	{
	  pEnd = pTmp + j;
	  cTmp = *pEnd;
	  *pEnd = '\0';
	  iCount = 0;
	  sih_lookup(ht, pTmp, &iCount);
	  iCount++;
	  sih_add(ht, strdup(pTmp), iCount);
	  *pEnd = cTmp;
	}
    }

  /* Now throw away counts that fail threshold */
  for (i=0; i<ht->nslots; i++)
    {
      if (!ht->slots[i].string) continue;
      len = strlen(ht->slots[i].string);
      assert (len > 0);
      assert (len <= MAX_N); 
      if ((double) ht->slots[i].intval < rgCurrentThresholds[len])
	{
	  int k, key;
	  /* Delete hash table entry */
	  free(ht->slots[i].string);
	  ht->slots[i].string = NULL;
	  ht->slots[i].intval = 0;
	  ht->nentries--;

	  k = (i+1) % ht->nslots;
	  for (; ; k++) {
	    char *pstr2;
	    int iVal;
	    k %= ht->nslots;
	    pstr2 = ht->slots[k].string;
	    iVal = ht->slots[k].intval;

	    if (!pstr2) break;
	    if ((double) iVal < rgCurrentThresholds[strlen(pstr2)]) continue;
	    SIH_KEY(pstr2, key);
	    if (k == key) continue;

	    else 
	      {
		ht->slots[k].string = NULL;
		ht->slots[k].intval = 0;
		ht->nentries--;
		sih_add(ht, pstr2, iVal);
	      }
	  }
	}
    }

  /* Increment the thresholds */
  for (i=1; i<=MAX_N; i++) rgCurrentThresholds[i] += rgBlockThresholds[i-1];

  fprintf(stderr, "[%d] ", ht->nentries);
}


void print_hash_table(lginfoptr lg)
{
  int i, len;
  sih_t *ht = lg->hashCounts;

  for (i=0; i<ht->nslots; i++)
    {
      if (!ht->slots[i].string) continue;
      len = strlen(ht->slots[i].string);
      assert (len > 0);
      assert (len <= MAX_N); 
      fprintf(stderr, "'%s' %d\n", 
	      ht->slots[i].string, ht->slots[i].intval);
    }
}


void print_train_info(lginfoptr lg)
{
  lginfoptr temp;

  fprintf(stderr, "\n\n%20s%30s%10s\n\n", "Language", "Training set size", "Clusters");
  for (temp = lg; temp; temp = temp -> next)
    {
      fprintf(stderr, "%20s%30d%10d\n", temp->name, temp->cChars, temp->hashCounts->nentries);
    }
}


void initlginfo(lginfoptr lg)
{
  int i;
  lg->hashCounts = sih_create(10000, 0.50, 1.1, FALSE);
  lg->cChars = 0;
  for (i=1; i<=MAX_N; i++) rgCurrentThresholds[i] = rgBlockThresholds[i-1];
}


lginfoptr do_training(char *filename)
{
  FILE *infile;
  lginfoptr langs, temp;
  char buff[200];

  langs = NULL;
  if ( ! (infile = fopen(filename, "r")))
    perror2("Do_training: Can't open file %s.\n", filename);

  while (fgets(buff, 200, infile))
    {
      temp = (lginfoptr) malloc(sizeof(struct lgstruct));
      temp -> next = langs;
      sscanf(buff, "%s %s", temp -> name, temp -> filename);
      fprintf(stderr, "\n\nProcessing %s:\n", temp->name);
      processfile(temp);
      langs = temp;
    }
  fclose(infile);
  print_train_info(langs);
  return langs;
}


double cosine_measure(int size, int a[], int b[])

{
  double dotprod = 0.0, sum2a = 0.0, sum2b = 0.0;
  int i;

  for (i = 0; i < size; i++)
    {
      dotprod += a[i] * b[i];
      sum2a += a[i] * a[i];
      sum2b += b[i] * b[i];
    }
#if DEBUG
  printf("dotprd, sum2a, sum2b, cos: %.2f %.2f %.2f %f\n",
	 dotprod, sum2a, sum2b, dotprod / sqrt(sum2a * sum2b));
#endif
  return dotprod / sqrt(sum2a * sum2b);
}


double calculate_score(lginfoptr lg, char *string)
{
  double sum = 0.0;
  int strcounts[CHARSET];
  int i, j, iVal = 0;
  double iSum=0.0;
  char *pstart, *pch, cTmp;
  int len = 0;
  int testlen;
  char szBuf[255], *pString, lastc='\0';

  szBuf[0] = '\0';

  /* Prepend a period, as if it were the start of a sentence */
  strcpy(szBuf, ". ");

  strcat(szBuf, string);

  if (isalpha(szBuf[0])) szBuf[0] = toupper(szBuf[0]);

  pString = szBuf;



  testlen = strlen(szBuf);
  if (testlen <= 0) return 0.0;

  for (i = 0; i < CHARSET; i++)
    strcounts[i] = 0;

  i =0;
  for ( ; *pString; pString++)
    {
      if (*pString == '\n') *pString=' ';
      /* if ((*pString == ' ') && (lastc == ' ')) continue;*/

      cTmp = *(pString+1);
      *(pString + 1) = '\0';
      if (i >= MAX_N) pstart = pString - MAX_N-1;
      else pstart  = pString;
      len = strlen(pstart);
      for (pch = pstart; len >= 1; pch++)
	{
	  if (sih_lookup(lg->hashCounts, pch, &iVal)) iSum+= rgWeights[len] +
	    ((double) iVal / (double) lg->cChars);
	  len--;
	}
      *(pString + 1) = cTmp;
      i++;
      lastc = *pString;
    }

  return iSum / (double) testlen;
}



lginfoptr calculate_best(lginfoptr lg, char *string)
{
  double sum = 0.0;
  int i, j, iVal = 0, iLang = 0, testlen;
  double iSum=0.0;
  char *pstart, *pch, cTmp;
  int len = 0;
  char szBuf[255], *pString;
  int logprobs[MAX_LANGS];
  int counts[MAX_LANGS], iTotalCount;
  lginfoptr lgbest = lg, temp; 
  int logBestScore = MIN_LOG;

  /* Prepend a period, as if it were the start of a sentence */
  strcpy(szBuf, ". ");

  strcat(szBuf, string);
  if (isalpha(szBuf[0])) szBuf[0] = toupper(szBuf[0]);

  pString = szBuf;

  testlen = strlen(szBuf);
  if (testlen <= 0) return lgbest;

  i=0;
  for ( ; *pString; pString++)
    {
      if (*pString == '\n') *pString=' ';
      cTmp = *(pString+1);
      *(pString + 1) = '\0';
      if (i >= MAX_N) pstart = pString - MAX_N-1;
      else pstart  = pString;
      len = strlen(pstart);
      for (pch = pstart; len >= 1; pch++)
	{
	  iTotalCount = 0;
	  iLang = 0;
	  for (temp = lg; temp; temp = temp -> next, iLang++)
	    {
	      if (sih_lookup(lg->hashCounts, pch, &iVal)) 
		{
		  counts[iLang] = iVal; 
		  iTotalCount += iVal;
		}
	      else
		{
		  counts[iLang] = 0;
		}
	    }

	  iLang = 0;
	  /* Second pass */
	  if (iTotalCount > 0) for (temp = lg; temp; temp = temp -> next, iLang++)
	    {
	      logprobs[iLang] += LOG(((double) counts[iLang] / (double) temp->cChars) /
				     ((double) iTotalCount));
	    }
	  len--;
	}
      
      *(pString + 1) = cTmp;
      i++;
    }

  /* Last pass */
  iLang = 0;
  for (temp = lg; temp; temp = temp -> next, iLang++)
    {
      if (logprobs[iLang] > logBestScore)
	{
	  lgbest = temp;
	  logBestScore = logprobs[iLang];
	}
    }
  return lgbest;
}



void do_testing(char *filename, lginfoptr lg)

{
  FILE *infile;
  lginfoptr best, temp;
  double val, bestval;
  char buff[200];
  
  if (!strcmp(filename, "-")) infile = stdin;
  else if ( ! (infile = fopen(filename, "r")))
    perror2("Do_testing: Can't open file %s.\n", filename);

  while (fgets(buff, 200, infile))
    {
      /* Evaluate string on each test language and print out best one */
      best = NULL;
      bestval = -1.0;

#if USE_PROBABILISTIC
      best = calculate_best(lg, buff);
#else
      for (temp = lg; temp; temp = temp -> next)
	{
#if DEBUG
	  fprintf(stderr, "Testing %s\n", temp -> name);
#endif
	  val = calculate_score(temp, buff);
	  if (infile == stdin)
	    fprintf(stderr, "Testing %s: %f\n", temp->name, val);
	  if (val > bestval)
	    {
	      best = temp;
	      bestval = val;
	    }
	}
#endif  /* USE_PROBABILISTIC */   
      fprintf(stdout, "%s\n", best -> name);
    }
  if (infile != stdin) fclose(infile);
}


main(int argc, char *argv[])

{
  lginfoptr langs;
  char szFile1[255], szFile2[255];
  int iArg;

  if (argc == 1)
    {
      strcpy(szFile1, "file1");
      strcpy(szFile2, "-");
    }
  else if (argc < 3)
    perror2("%s: trainingfile testfile\n", argv[0]);
  else
    {
      strcpy(szFile1, argv[1]);
      strcpy(szFile2, argv[2]);
    }

  /* Load the data */
  if ((langs = do_training(szFile1)) == NULL)
    perror2("I didn't find any training info in file %s.\n", szFile1);

  /* Do testing */
  do_testing(szFile2, langs);
}


#if 0
void sih_delete(sih_t *ht, char *string)
{
  int k, key;
}
#endif
